use crate::common::Section;
use crate::common::SectionInfo;
use crate::error::Error;

const SI_TEXT: usize = 0;
const SI_RODATA: usize = 1;
const SI_DATA: usize = 2;
const SI_BSS: usize = 3;

pub enum SegmentType {
    Text = SI_TEXT as isize,
    Data = SI_DATA as isize,
    Bss = SI_BSS as isize,
}

const SEGMENT_NAMES: [&str; 4] = [".text", ".rodata", ".data", ".bss"];

pub struct SegmentData<'data>([Vec<SectionInfo<'data>>; 4]);

impl<'data> SegmentData<'data> {
    pub fn new() -> Self {
        Self([Vec::new(), Vec::new(), Vec::new(), Vec::new()])
    }

    pub fn append_section(&mut self, section: Section<'data>) -> Result<(), Error> {
        match section {
            Section::Text(si) => self.0[SI_TEXT].push(si),
            Section::Data(si, false) => self.0[SI_DATA].push(si),
            Section::Data(si, true) => self.0[SI_RODATA].push(si),
            Section::Bss(si) => self.0[SI_BSS].push(si),
        }

        Ok(())
    }

    pub fn iter_mut(
        &mut self,
    ) -> impl Iterator<Item = (&'static str, &mut Vec<SectionInfo<'data>>)> {
        self.0
            .iter_mut()
            .enumerate()
            .map(|(i, s)| (SEGMENT_NAMES[i], s))
    }

    // .text + .rodata
    pub fn program_size(&self) -> u64 {
        let text_iter = self.0[SI_TEXT].iter();
        let rodata_iter = self.0[SI_RODATA].iter();

        let mut result = 0u64;
        for section in text_iter.chain(rodata_iter) {
            result += section.size
        }

        result
    }

    // data
    pub fn data_size(&self) -> u64 {
        let mut result = 0u64;
        for section in self.0[SI_DATA].iter() {
            result += section.size
        }

        result
    }

    // bss
    pub fn bss_size(&self) -> u64 {
        let mut result = 0u64;
        for section in self.0[SI_BSS].iter() {
            result += section.size
        }

        result
    }
}

impl Default for SegmentData<'_> {
    fn default() -> Self {
        Self::new()
    }
}
